@@ -46,7 +46,7 @@ our @EXPORT = qw( process_tos
 		  generate_matrix
 		  );
 our @EXPORT_OK = qw( initialize );
-our $VERSION = '4.5_18';
+our $VERSION = '4.5_21';
 
 our $family;
 
@@ -679,7 +679,13 @@ sub process_stoppedrules() {
     my $result;
 
     if ( my $fn = open_file 'stoppedrules' , 1, 1 ) {
-	first_entry "$doing $fn...";
+	first_entry sub() {
+	    progress_message2("$doing $fn...");
+	    unless ( $config{ADMINISABSENTMINDED} ) {
+		warning_message("Entries in the routestopped file are processed as if ADMINISABSENTMINDED=Yes");
+		$config{ADMINISABSENTMINDED} = 'Yes';
+	    }
+	};
 
 	while ( read_a_line( NORMAL_READ ) ) {
 
@@ -832,35 +838,38 @@ sub add_common_rules ( $ ) {
     }
 
     for $interface ( all_real_interfaces ) {
-	ensure_chain( 'filter', $_ ) for first_chains( $interface ), output_chain( $interface ), option_chains( $interface ), output_option_chain( $interface );
+	ensure_chain( 'filter', $_ )->{origin} = interface_origin( $interface )
+	    for first_chains( $interface ), output_chain( $interface ), option_chains( $interface ), output_option_chain( $interface );
 
 	my $interfaceref = find_interface $interface;
 
-	unless ( $interfaceref->{options}{ignore} & NO_SFILTER || $interfaceref->{options}{rpfilter} || $interfaceref->{physical} eq 'lo' ) {
+	unless ( $interfaceref->{physical} eq 'lo' ) {
+	    unless ( $interfaceref->{options}{ignore} & NO_SFILTER || $interfaceref->{options}{rpfilter} ) {
 
-	    my @filters = @{$interfaceref->{filter}};
+		my @filters = @{$interfaceref->{filter}};
 
-	    $chainref = $filter_table->{forward_option_chain $interface};
+		$chainref = $filter_table->{forward_option_chain $interface};
 
-	    if ( @filters ) {
-		add_ijump( $chainref , @ipsec ? 'j' : 'g' => $target1, imatch_source_net( $_ ), @ipsec ), $chainref->{filtered}++ for @filters;
-	    } elsif ( $interfaceref->{bridge} eq $interface ) {
-		add_ijump( $chainref , @ipsec ? 'j' : 'g' => $target1, imatch_dest_dev( $interface ), @ipsec ), $chainref->{filtered}++
-		    unless( $config{ROUTE_FILTER} eq 'on' ||
-			    $interfaceref->{options}{routeback} ||
-			    $interfaceref->{options}{routefilter} ||
-			    $interfaceref->{physical} eq '+' );
-	    }
+		if ( @filters ) {
+		    add_ijump( $chainref , @ipsec ? 'j' : 'g' => $target1, imatch_source_net( $_ ), @ipsec ), $chainref->{filtered}++ for @filters;
+		} elsif ( $interfaceref->{bridge} eq $interface ) {
+		    add_ijump( $chainref , @ipsec ? 'j' : 'g' => $target1, imatch_dest_dev( $interface ), @ipsec ), $chainref->{filtered}++
+			unless( $config{ROUTE_FILTER} eq 'on' ||
+				$interfaceref->{options}{routeback} ||
+				$interfaceref->{options}{routefilter} ||
+				$interfaceref->{physical} eq '+' );
+		}
 
 
-	    if ( @filters ) {
-		$chainref = $filter_table->{input_option_chain $interface};
-		add_ijump( $chainref , g => $target, imatch_source_net( $_ ), @ipsec ), $chainref->{filtered}++ for @filters;
+		if ( @filters ) {
+		    $chainref = $filter_table->{input_option_chain $interface};
+		    add_ijump( $chainref , g => $target, imatch_source_net( $_ ), @ipsec ), $chainref->{filtered}++ for @filters;
+		}
 	    }
 
 	    for ( option_chains( $interface ) ) {
 		add_ijump( $filter_table->{$_}, j => $dynamicref, @state ) if $dynamicref;
-		add_ijump( $filter_table->{$_}, j => 'ACCEPT', state_imatch $faststate ) if $config{FASTACCEPT};
+		add_ijump( $filter_table->{$_}, j => 'ACCEPT', state_imatch $faststate )->{comment} = '' if $config{FASTACCEPT};
 	    }
 	}
     }
@@ -926,14 +935,13 @@ sub add_common_rules ( $ ) {
 	if ( supplied $config{SMURF_LOG_LEVEL} ) {
 	    my $smurfref = new_chain( 'filter', 'smurflog' );
 
-	    log_rule_limit( $config{SMURF_LOG_LEVEL},
-			    $smurfref,
-			    'smurfs' ,
-			    'DROP',
-			    $globals{LOGLIMIT},
-			    '',
-			    'add',
-			    '' );
+	    log_irule_limit( $config{SMURF_LOG_LEVEL},
+			     $smurfref,
+			     'smurfs' ,
+			     'DROP',
+			     $globals{LOGILIMIT},
+			     '',
+			     'add' );
 	    add_ijump( $smurfref, j => 'AUDIT', targetopts => '--type drop' ) if $smurfdest eq 'A_DROP';
 	    add_ijump( $smurfref, j => 'DROP' );
 
@@ -983,42 +991,44 @@ sub add_common_rules ( $ ) {
 	}
     }
 
-    if ( have_capability( 'ADDRTYPE' ) ) {
-	add_ijump $rejectref , j => 'DROP' , addrtype => '--src-type BROADCAST';
-    } else {
-	if ( $family == F_IPV4 ) {
-	    add_commands $rejectref, 'for address in $ALL_BCASTS; do';
+    unless ( $config{REJECT_ACTION} ) {
+	if ( have_capability( 'ADDRTYPE' ) ) {
+	    add_ijump $rejectref , j => 'DROP' , addrtype => '--src-type BROADCAST';
 	} else {
-	    add_commands $rejectref, 'for address in $ALL_ACASTS; do';
+	    if ( $family == F_IPV4 ) {
+		add_commands $rejectref, 'for address in $ALL_BCASTS; do';
+	    } else {
+		add_commands $rejectref, 'for address in $ALL_ACASTS; do';
+	    }
+
+	    incr_cmd_level $rejectref;
+	    add_ijump $rejectref, j => 'DROP', d => '$address';
+	    decr_cmd_level $rejectref;
+	    add_commands $rejectref, 'done';
 	}
 
-	incr_cmd_level $rejectref;
-	add_ijump $rejectref, j => 'DROP', d => '$address';
-	decr_cmd_level $rejectref;
-	add_commands $rejectref, 'done';
-    }
-
-    if ( $family == F_IPV4 ) {
-	add_ijump $rejectref , j => 'DROP', s => '224.0.0.0/4';
-    } else {
-	add_ijump $rejectref , j => 'DROP', s => IPv6_MULTICAST;
-    }
-
-    add_ijump $rejectref , j => 'DROP', p => 2;
-    add_ijump $rejectref , j => 'REJECT', targetopts => '--reject-with tcp-reset', p => 6;
-
-    if ( have_capability( 'ENHANCED_REJECT' ) ) {
-	add_ijump $rejectref , j => 'REJECT', p => 17;
-
 	if ( $family == F_IPV4 ) {
-	    add_ijump $rejectref, j => 'REJECT --reject-with icmp-host-unreachable', p => 1;
-	    add_ijump $rejectref, j => 'REJECT --reject-with icmp-host-prohibited';
+	    add_ijump $rejectref , j => 'DROP', s => '224.0.0.0/4';
 	} else {
-	    add_ijump $rejectref, j => 'REJECT --reject-with icmp6-addr-unreachable', p => 58;
-	    add_ijump $rejectref, j => 'REJECT --reject-with icmp6-adm-prohibited';
+	    add_ijump $rejectref , j => 'DROP', s => IPv6_MULTICAST;
+	}
+
+	add_ijump $rejectref , j => 'DROP', p => 2;
+	add_ijump $rejectref , j => 'REJECT', targetopts => '--reject-with tcp-reset', p => 6;
+
+	if ( have_capability( 'ENHANCED_REJECT' ) ) {
+	    add_ijump $rejectref , j => 'REJECT', p => 17;
+
+	    if ( $family == F_IPV4 ) {
+		add_ijump $rejectref, j => 'REJECT --reject-with icmp-host-unreachable', p => 1;
+		add_ijump $rejectref, j => 'REJECT --reject-with icmp-host-prohibited';
+	    } else {
+		add_ijump $rejectref, j => 'REJECT --reject-with icmp6-addr-unreachable', p => 58;
+		add_ijump $rejectref, j => 'REJECT --reject-with icmp6-adm-prohibited';
+	    }
+	} else {
+	    add_ijump $rejectref , j => 'REJECT';
 	}
-    } else {
-	add_ijump $rejectref , j => 'REJECT';
     }
 
     $list = find_interfaces_by_option 'dhcp';
@@ -1334,7 +1344,7 @@ sub setup_mac_lists( $ ) {
 
 	    run_user_exit2( 'maclog', $chainref );
 
-	    log_rule_limit $level, $chainref , $chain , $disposition, '', '', 'add', '' if $level ne '';
+	    log_irule_limit $level, $chainref , $chain , $disposition, [], '', 'add' if $level ne '';
 	    add_ijump $chainref, j => $target;
 	}
     }
@@ -2283,7 +2293,7 @@ sub generate_matrix() {
 		    '' ,
 		    '' ,
 		    'insert' ,
-		    "$globals{STATEMATCH} NEW ";
+		    state_match('NEW');
 	    }
 	}
     }
@@ -2510,9 +2520,9 @@ EOF
 	  "restore_default_route $config{USE_DEFAULT_RT}"
 	  );
 
-    my @chains = $config{ADMINISABSENTMINDED} ? qw/INPUT FORWARD/ : qw/INPUT OUTPUT FORWARD/;
-
-    add_ijump $filter_table ->{$_}, j => 'ACCEPT', state_imatch 'ESTABLISHED,RELATED' for @chains;
+    if ( $config{ADMINISABSENTMINDED} ) {
+	add_ijump $filter_table ->{$_}, j => 'ACCEPT', state_imatch 'ESTABLISHED,RELATED' for qw/INPUT FORWARD/;
+    }
 
     if ( $family == F_IPV6 ) {
 	add_ijump $input, j => 'ACCEPT', s => IPv6_LINKLOCAL;
