@@ -186,7 +186,7 @@ sub setup_route_marking() {
 sub copy_table( $$$ ) {
     my ( $duplicate, $number, $realm ) = @_;
 
-    my $filter = $family == F_IPV6 ? q(fgrep -v ' cache ' | sed 's/ via :: / /' | ) : '';
+    my $filter = $family == F_IPV6 ? q(grep -vF ' cache ' | sed 's/ via :: / /' | ) : '';
 
     emit '';
 
@@ -223,7 +223,7 @@ sub copy_table( $$$ ) {
 sub copy_and_edit_table( $$$$$ ) {
     my ( $duplicate, $number, $id, $copy, $realm) = @_;
 
-    my $filter = $family == F_IPV6 ? q(fgrep -v ' cache ' | sed 's/ via :: / /' | ) : '';
+    my $filter = $family == F_IPV6 ? q(grep -vF ' cache ' | sed 's/ via :: / /' | ) : '';
     my %copied;
     my @copy;
     my @bup_copy;
@@ -474,14 +474,14 @@ sub process_a_provider( $ ) {
 	$gateway = '';
     }
 
-    my ( $loose, $track, $balance, $default, $default_balance, $optional, $mtu, $tproxy, $local, $load, $what );
+    my ( $loose, $track, $balance, $default, $default_balance, $optional, $mtu, $tproxy, $local, $load, $what, $hostroute );
 
     if ( $pseudo ) {	
-	( $loose, $track,                   $balance , $default, $default_balance,                $optional,                           $mtu, $tproxy , $local, $load, $what ) =
-	( 0,      0                       , 0 ,        0,        0,                               1                                  , ''  , 0       , 0,      0,     'interface');
+	( $loose, $track,                   $balance , $default, $default_balance,                $optional,                           $mtu, $tproxy , $local, $load, $what ,      $hostroute ) =
+	( 0,      0                       , 0 ,        0,        0,                               1                                  , ''  , 0       , 0,      0,     'interface', 0);
     } else {
-	( $loose, $track,                   $balance , $default, $default_balance,                $optional,                           $mtu, $tproxy , $local, $load, $what )=
-	( 0,      $config{TRACK_PROVIDERS}, 0 ,        0,        $config{USE_DEFAULT_RT} ? 1 : 0, interface_is_optional( $interface ), ''  , 0       , 0,      0,     'provider');
+	( $loose, $track,                   $balance , $default, $default_balance,                $optional,                           $mtu, $tproxy , $local, $load, $what      , $hostroute )=
+	( 0,      $config{TRACK_PROVIDERS}, 0 ,        0,        $config{USE_DEFAULT_RT} ? 1 : 0, interface_is_optional( $interface ), ''  , 0       , 0,      0,     'provider',  1);
     }
 
     unless ( $options eq '-' ) {
@@ -535,6 +535,10 @@ sub process_a_provider( $ ) {
 		$noautosrc = 0;
 	    } elsif ( $option eq 'noautosrc' ) {
 		$noautosrc = 1;
+	    } elsif ( $option eq 'hostroute' ) {
+		$hostroute = 1;
+	    } elsif ( $option eq 'nohostroute' ) {
+		$hostroute = 0;
 	    } else {
 		fatal_error "Invalid option ($option)";
 	    }
@@ -642,6 +646,7 @@ sub process_a_provider( $ ) {
 			   load        => $load ,
 			   pseudo      => $pseudo ,
 			   what        => $what ,
+			   hostroute   => $hostroute ,
 			   rules       => [] ,
 			   routes      => [] ,
 			   routedests  => {} ,
@@ -718,6 +723,7 @@ sub add_a_provider( $$ ) {
     my $pseudo      = $providerref->{pseudo};
     my $what        = $providerref->{what};
     my $label       = $pseudo ? 'Optional Interface' : 'Provider';
+    my $hostroute   = $providerref->{hostroute};
 
     my $dev         = var_base $physical;
     my $base        = uc $dev;
@@ -798,13 +804,16 @@ CEOF
 
     if ( $gateway ) {
 	$address = get_interface_address $interface unless $address;
-	if ( $family == F_IPV4 ) {
-	    emit "run_ip route replace $gateway src $address dev $physical ${mtu}";
-	    emit "run_ip route replace $gateway src $address dev $physical ${mtu}table $id $realm";
-	} else {
-	    emit "qt \$IP -6 route add $gateway src $address dev $physical ${mtu}";
-	    emit "qt \$IP -6 route del $gateway src $address dev $physical ${mtu}table $id $realm";
-	    emit "run_ip route add $gateway src $address dev $physical ${mtu}table $id $realm";
+
+	if ( $hostroute ) {
+	    if ( $family == F_IPV4 ) {
+		emit "run_ip route replace $gateway src $address dev $physical ${mtu}";
+		emit "run_ip route replace $gateway src $address dev $physical ${mtu}table $id $realm";
+	    } else {
+		emit "qt \$IP -6 route add $gateway src $address dev $physical ${mtu}";
+		emit "qt \$IP -6 route del $gateway src $address dev $physical ${mtu}table $id $realm";
+		emit "run_ip route add $gateway src $address dev $physical ${mtu}table $id $realm";
+	    }
 	}
 
 	emit "run_ip route add default via $gateway src $address dev $physical ${mtu}table $id $realm";
@@ -819,7 +828,7 @@ CEOF
 	emit '';
 	if ( $gateway ) {
 	    if ( $family == F_IPV4 ) {
-		emit qq(run_ip route replace $gateway/32 dev $physical table $id);
+		emit qq(run_ip route replace $gateway/32 dev $physical table $id) if $hostroute;
 		emit qq(run_ip route add default via $gateway src $address dev $physical table $id metric $number);
 	    } else {
 		emit qq(qt \$IP -6 route del default via $gateway src $address dev $physical table $id metric $number);
@@ -854,12 +863,12 @@ CEOF
 	    if ( $shared ) {
 		emit  "qt \$IP -$family rule del from $address" if $config{DELETE_THEN_ADD};
 		emit( "run_ip rule add from $address pref 20000 table $id" ,
-		      "echo \"\$IP -$family rule del from $address > /dev/null 2>&1\" >> \${VARDIR}/undo_${table}_routing" );
+		      "echo \"\$IP -$family rule del from $address pref 20000> /dev/null 2>&1\" >> \${VARDIR}/undo_${table}_routing" );
 	    } elsif ( ! $pseudo ) {
 		emit  ( "find_interface_addresses $physical | while read address; do" );
 		emit  ( "    qt \$IP -$family rule del from \$address" ) if $config{DELETE_THEN_ADD};
 		emit  ( "    run_ip rule add from \$address pref 20000 table $id",
-			"    echo \"\$IP -$family rule del from \$address > /dev/null 2>&1\" >> \${VARDIR}/undo_${table}_routing",
+			"    echo \"\$IP -$family rule del from \$address pref 20000 > /dev/null 2>&1\" >> \${VARDIR}/undo_${table}_routing",
 			'    rulenum=$(($rulenum + 1))',
 			'done'
 		      );
@@ -1989,18 +1998,18 @@ sub handle_stickiness( $ ) {
 
 		for my $chainref ( $stickyref, $setstickyref ) {
 		    if ( $chainref->{name} eq 'sticky' ) {
-			$rule1 = clone_rule( $_ );
+			$rule1 = clone_irule( $_ );
 
 			set_rule_target( $rule1, 'MARK',   "--set-mark $mark" );
-			set_rule_option( $rule1, 'recent', "--name $list --update --seconds 300" );
+			set_rule_option( $rule1, 'recent', "--name $list --update --seconds 300 --reap" );
 
-			$rule2 = clone_rule( $_ );
+			$rule2 = clone_irule( $_ );
 
 			clear_rule_target( $rule2 );
 			set_rule_option( $rule2, 'mark',   "--mark 0\/$mask" );
 			set_rule_option( $rule2, 'recent', "--name $list --remove" );
 		    } else {
-			$rule1 = clone_rule( $_ );
+			$rule1 = clone_irule( $_ );
 
 			clear_rule_target( $rule1 );
 			set_rule_option( $rule1, 'mark',   "--mark $mark\/$mask" );
@@ -2024,18 +2033,18 @@ sub handle_stickiness( $ ) {
 
 		for my $chainref ( $stickoref, $setstickoref ) {
 		    if ( $chainref->{name} eq 'sticko' ) {
-			$rule1 = clone_rule $_;
+			$rule1 = clone_irule $_;
 
 			set_rule_target( $rule1, 'MARK',   "--set-mark $mark" );
-			set_rule_option( $rule1, 'recent', " --name $list --rdest --update --seconds 300" );
+			set_rule_option( $rule1, 'recent', " --name $list --rdest --update --seconds 300 --reap" );
 
-			$rule2 = clone_rule $_;
+			$rule2 = clone_irule $_;
 
 			clear_rule_target( $rule2 );
 			set_rule_option  ( $rule2, 'mark',   "--mark 0\/$mask" );
 			set_rule_option  ( $rule2, 'recent', "--name $list --rdest --remove" );
 		    } else {
-			$rule1 = clone_rule $_;
+			$rule1 = clone_irule $_;
 
 			clear_rule_target( $rule1 );
 			set_rule_option  ( $rule1, 'mark',   "--mark $mark" );
