@@ -54,10 +54,11 @@ our @EXPORT = qw(
 		  perl_action_helper
 		  perl_action_tcp_helper
 		  check_state
+                  process_reject_action
 	       );
 
 our @EXPORT_OK = qw( initialize process_rule );
-our $VERSION = '4.5_18';
+our $VERSION = '4.5_21';
 #
 # Globals are documented in the initialize() function
 #
@@ -383,6 +384,7 @@ sub set_policy_chain($$$$$$)
 
 	$chainref1->{policy} = $policy;
 	$chainref1->{policypair} = [ $source, $dest ];
+	$chainref1->{origin} = $chainref->{origin};
     }
 }
 
@@ -573,6 +575,8 @@ sub process_a_policy() {
     my $chainref1 = $usedactions{$default};
     $chainref->{default} = $chainref1 ? $chainref1->{name} : $default;
 
+    $chainref->{origin} = shortlineinfo('');
+
     if ( $clientwild ) {
 	if ( $serverwild ) {
 	    for my $zone ( @zonelist ) {
@@ -880,11 +884,13 @@ sub complete_standard_chain ( $$$$ ) {
     $policychainref = $filter_table->{$ruleschainref->{policychain}} if $ruleschainref;
 
     if ( $policychainref ) {
-	( $policy, $loglevel, $defaultaction ) = @{$policychainref}{'policy', 'loglevel', 'default' }
+	( $policy, $loglevel, $defaultaction ) = @{$policychainref}{'policy', 'loglevel', 'default' };
+	$stdchainref->{origin} = $policychainref->{origin};
     } elsif ( $defaultaction !~ /:/ ) {
 	$defaultaction = join(":", $defaultaction, 'none', '', '' );
     }
 
+
     policy_rules $stdchainref , $policy , $loglevel, $defaultaction, 0;
 }
 
@@ -901,14 +907,13 @@ sub setup_syn_flood_chains() {
 		    new_chain 'filter' , syn_flood_chain $chainref :
 		    new_chain( 'filter' , '@' . $chainref->{name} );
 	    add_rule $synchainref , "${limit}-j RETURN";
-	    log_rule_limit( $level ,
-			    $synchainref ,
-			    $chainref->{name} ,
-			    'DROP',
-			    $globals{LOGLIMIT} || '-m limit --limit 5/min --limit-burst 5 ' ,
+	    log_irule_limit( $level ,
+			     $synchainref ,
+			     $chainref->{name} ,
+			     'DROP',
+			     @{$globals{LOGILIMIT}} ? $globals{LOGILIMIT} : [ limit => "--limit 5/min --limit-burst 5" ] ,
 			    '' ,
-			    'add' ,
-			    '' )
+			    'add' )
 		if $level ne '';
 	    add_ijump $synchainref, j => 'DROP';
 	}
@@ -1471,11 +1476,11 @@ sub dropBcast( $$$$ ) {
 
     if ( have_capability( 'ADDRTYPE' ) ) {
 	if ( $level ne '' ) {
-	    log_rule_limit $level, $chainref, 'dropBcast' , 'DROP', '', $tag, 'add', ' -m addrtype --dst-type BROADCAST ';
+	    log_irule_limit( $level, $chainref, 'dropBcast' , 'DROP', [], $tag, 'add', addrtype => '--dst-type BROADCAST' );
 	    if ( $family == F_IPV4 ) {
-		log_rule_limit $level, $chainref, 'dropBcast' , 'DROP', '', $tag, 'add', ' -d 224.0.0.0/4 ';
+		log_irule_limit( $level, $chainref, 'dropBcast' , 'DROP', [], $tag, 'add', d => '224.0.0.0/4' );
 	    } else {
-		log_rule_limit $level, $chainref, 'dropBcast' , 'DROP', '', $tag, 'add', join( ' ', ' -d' , IPv6_MULTICAST , '-j DROP ' );
+		log_irule_limit( $level, $chainref, 'dropBcast' , 'DROP', [], $tag, 'add', d => IPv6_MULTICAST );
 	    }
 	}
 
@@ -1488,17 +1493,17 @@ sub dropBcast( $$$$ ) {
 	}
 
 	incr_cmd_level $chainref;
-	log_rule_limit $level, $chainref, 'dropBcast' , 'DROP', '', $tag, 'add', ' -d $address ' if $level ne '';
+	log_irule_limit( $level, $chainref, 'dropBcast' , 'DROP', [], $tag, 'add', d => '$address' ) if $level ne '';
 	add_ijump $chainref, j => $target, d => '$address';
 	decr_cmd_level $chainref;
 	add_commands $chainref, 'done';
     }
 
     if ( $family == F_IPV4 ) {
-	log_rule_limit $level, $chainref, 'dropBcast' , 'DROP', '', $tag, 'add', ' -d 224.0.0.0/4 ' if $level ne '';
+	log_irule_limit $level, $chainref, 'dropBcast' , 'DROP', [], $tag, 'add', d => '224.0.0.0/4' if $level ne '';
 	add_ijump $chainref, j => $target, d => '224.0.0.0/4';
     } else {
-	log_rule_limit $level, $chainref, 'dropBcast' , 'DROP', '', $tag, 'add', join( ' ', ' -d' , IPv6_MULTICAST . ' ' ) if $level ne '';
+	log_irule_limit( $level, $chainref, 'dropBcast' , 'DROP', [], $tag, 'add', d => IPv6_MULTICAST ) if $level ne '';
 	add_ijump $chainref, j => $target, d => IPv6_MULTICAST;
     }
 }
@@ -1510,8 +1515,8 @@ sub allowBcast( $$$$ ) {
 
     if ( $family == F_IPV4 && have_capability( 'ADDRTYPE' ) ) {
 	if ( $level ne '' ) {
-	    log_rule_limit $level, $chainref, 'allowBcast' , 'ACCEPT', '', $tag, 'add', ' -m addrtype --dst-type BROADCAST ';
-	    log_rule_limit $level, $chainref, 'allowBcast' , 'ACCEPT', '', $tag, 'add', ' -d 224.0.0.0/4 ';
+	    log_irule_limit( $level, $chainref, 'allowBcast' , 'ACCEPT', [], $tag, 'add', addrtype => '--dst-type BROADCAST' );
+	    log_irule_limit( $level, $chainref, 'allowBcast' , 'ACCEPT', [], $tag, 'add', d => '224.0.0.0/4' );
 	}
 
 	add_ijump $chainref, j => $target, addrtype => '--dst-type BROADCAST';
@@ -1523,17 +1528,17 @@ sub allowBcast( $$$$ ) {
 	}
 
 	incr_cmd_level $chainref;
-	log_rule_limit $level, $chainref, 'allowBcast' , 'ACCEPT', '', $tag, 'add', ' -d $address ' if $level ne '';
+	log_irule_limit( $level, $chainref, 'allowBcast' , 'ACCEPT', [], $tag, 'add', d => '$address' ) if $level ne '';
 	add_ijump $chainref, j => $target, d => '$address';
 	decr_cmd_level $chainref;
 	add_commands $chainref, 'done';
     }
 
     if ( $family == F_IPV4 ) {
-	log_rule_limit $level, $chainref, 'allowBcast' , 'ACCEPT', '', $tag, 'add', ' -d 224.0.0.0/4 ' if $level ne '';
+	log_irule_limit( $level, $chainref, 'allowBcast' , 'ACCEPT', [], $tag, 'add', d => '224.0.0.0/4' ) if $level ne '';
 	add_ijump $chainref, j => $target, d => '224.0.0.0/4';
     } else {
-	log_rule_limit $level, $chainref, 'allowBcast' , 'ACCEPT', '', $tag, 'add', ' -d ' . IPv6_MULTICAST . ' ' if $level ne '';
+	log_irule_limit( $level, $chainref, 'allowBcast' , 'ACCEPT', '', $tag, 'add',  d => IPv6_MULTICAST ) if $level ne '';
 	add_ijump $chainref, j => $target, d => IPv6_MULTICAST;
     }
 }
@@ -1543,7 +1548,7 @@ sub dropNotSyn ( $$$$ ) {
 
     my $target = require_audit( 'DROP', $audit );
 
-    log_rule_limit $level, $chainref, 'dropNotSyn' , 'DROP', '', $tag, 'add', '-p 6 ! --syn ' if $level ne '';
+    log_irule_limit( $level, $chainref, 'dropNotSyn' , 'DROP', [], $tag, 'add', p => '6 ! --syn' ) if $level ne '';
     add_ijump $chainref , j => $target, p => '6 ! --syn';
 }
 
@@ -1558,7 +1563,7 @@ sub rejNotSyn ( $$$$ ) {
 	$target = require_audit( 'REJECT' , $audit );
     }
 
-    log_rule_limit $level, $chainref, 'rejNotSyn' , 'REJECT', '', $tag, 'add', '-p 6 ! --syn ' if $level ne '';
+    log_irule_limit( $level, $chainref, 'rejNotSyn' , 'REJECT', [], $tag, 'add', p => '6 ! --syn' ) if $level ne '';
     add_ijump $chainref , j => $target, p => '6 ! --syn';
 }
 
@@ -1574,8 +1579,8 @@ sub allowinUPnP ( $$$$ ) {
     my $target = require_audit( 'ACCEPT', $audit );
 
     if ( $level ne '' ) {
-	log_rule_limit $level, $chainref, 'allowinUPnP' , 'ACCEPT', '', $tag, 'add', '-p 17 --dport 1900 ';
-	log_rule_limit $level, $chainref, 'allowinUPnP' , 'ACCEPT', '', $tag, 'add', '-p 6 --dport 49152 ';
+	log_irule_limit( $level, $chainref, 'allowinUPnP' , 'ACCEPT', [], $tag, 'add', p => '17 --dport 1900' );
+	log_irule_limit( $level, $chainref, 'allowinUPnP' , 'ACCEPT', [], $tag, 'add', p => '6 --dport 49152' );
     }
 
     add_ijump $chainref, j => $target, p => '17 --dport 1900';
@@ -1606,11 +1611,13 @@ sub Limit( $$$$ ) {
 
     require_capability( 'RECENT_MATCH' , 'Limit rules' , '' );
 
+    warning_message "The Limit action is deprecated in favor of per-IP rate limiting using the RATE LIMIT column";
+
     add_irule $chainref, recent => "--name $set --set";
 
     if ( $level ne '' ) {
 	my $xchainref = new_chain 'filter' , "$chainref->{name}%";
-	log_rule_limit $level, $xchainref, $param[0], 'DROP', '', $tag, 'add', '';
+	log_irule_limit( $level, $xchainref, $param[0], 'DROP', [], $tag, 'add' );
 	add_ijump $xchainref, j => 'DROP';
 	add_ijump $chainref,  j => $xchainref, recent => "--name $set --update --seconds $param[2] --hitcount $count";
     } else {
@@ -1629,91 +1636,6 @@ my %builtinops = ( 'dropBcast'      => \&dropBcast,
 		   'Limit'          => \&Limit,
 		 );
 
-#
-# This function is called prior to processing of the policy file. It:
-#
-# - Adds the builtin actions to the target table
-# - Reads actions.std and actions (in that order) and for each entry:
-#   o Adds the action to the target table
-#   o Verifies that the corresponding action file exists
-#
-
-sub process_actions() {
-
-    progress_message2 "Locating Action Files...";
-    #
-    # Add built-in actions to the target table and create those actions
-    #
-    $targets{$_} = new_action( $_ , ACTION + BUILTIN, 1, 0 ) for @builtins;
-
-    for my $file ( qw/actions.std actions/ ) {
-	open_file( $file, 2 );
-
-	while ( read_a_line( NORMAL_READ ) ) {
-	    my ( $action, $options ) = split_line 'action file' , { action => 0, options => 1 };
-
-	    my $type     = ACTION;
-	    my $noinline = 0;
-	    my $nolog    = 0;
-	    my $builtin  = 0;
-
-	    if ( $action =~ /:/ ) {
-		warning_message 'Default Actions are now specified in /etc/shorewall/shorewall.conf';
-		$action =~ s/:.*$//;
-	    }
-
-	    fatal_error "Invalid Action Name ($action)" unless $action =~ /^[a-zA-Z][\w-]*$/;
-
-	    if ( $options ne '-' ) {
-		for ( split_list( $options, 'option' ) ) {
-		    if ( $_ eq 'inline' ) {
-			$type = INLINE;
-		    } elsif ( $_ eq 'noinline' ) {
-			$noinline = 1;
-		    } elsif ( $_ eq 'nolog' ) {
-			$nolog = 1;
-		    } elsif ( $_ eq 'builtin' ) {
-			$builtin = 1;
-		    } else {
-			fatal_error "Invalid option ($_)";
-		    }
-		}
-	    }
-
-	    fatal_error "Conflicting OPTIONS ($options)" if $noinline && $type == INLINE; 
-
-	    if ( my $actiontype = $targets{$action} ) {
-		if ( ( $actiontype & ACTION ) && ( $type == INLINE ) ) {
-		    if ( $actions{$action}->{noinline} ) {
-			warning_message "'inline' option ignored on action $action -- that action may not be in-lined";
-			next;
-		    }
-		    
-		    delete $actions{$action};
-		    delete $targets{$action};
-		} else {
-		    warning_message "Duplicate Action Name ($action) Ignored" unless $actiontype & ( ACTION | INLINE );
-		    next;
-		}
-	    }
-
-	    if ( $builtin ) {
-		$targets{$action}         = USERBUILTIN + OPTIONS;
-		$builtin_target{$action}  = 1;
-	    } else {
-		new_action $action, $type, $noinline, $nolog;
-
-		my $actionfile = find_file( "action.$action" );
-
-		fatal_error "Missing Action File ($actionfile)" unless -f $actionfile;
-
-		$inlines{$action} = { file => $actionfile, nolog => $nolog } if $type == INLINE;
-	    }
-	}
-    }
-
-}
-
 sub process_rule ( $$$$$$$$$$$$$$$$$$$ );
 
 #
@@ -1804,6 +1726,96 @@ sub process_action($$) {
     pop_action_params( $oldparms );
 }
 
+#
+# This function is called prior to processing of the policy file. It:
+#
+# - Adds the builtin actions to the target table
+# - Reads actions.std and actions (in that order) and for each entry:
+#   o Adds the action to the target table
+#   o Verifies that the corresponding action file exists
+#
+
+sub process_actions() {
+
+    progress_message2 "Locating Action Files...";
+    #
+    # Add built-in actions to the target table and create those actions
+    #
+    $targets{$_} = new_action( $_ , ACTION + BUILTIN, 1, 0 ) for @builtins;
+
+    for my $file ( qw/actions.std actions/ ) {
+	open_file( $file, 2 );
+
+	while ( read_a_line( NORMAL_READ ) ) {
+	    my ( $action, $options ) = split_line 'action file' , { action => 0, options => 1 };
+
+	    my $type     = ( $action eq $config{REJECT_ACTION} ? INLINE : ACTION );
+	    my $noinline = 0;
+	    my $nolog    = ( $type == INLINE ) || 0;
+	    my $builtin  = 0;
+
+	    if ( $action =~ /:/ ) {
+		warning_message 'Default Actions are now specified in /etc/shorewall/shorewall.conf';
+		$action =~ s/:.*$//;
+	    }
+
+	    fatal_error "Invalid Action Name ($action)" unless $action =~ /^[a-zA-Z][\w-]*$/;
+
+	    if ( $options ne '-' ) {
+		for ( split_list( $options, 'option' ) ) {
+		    if ( $_ eq 'inline' ) {
+			$type = INLINE;
+		    } elsif ( $_ eq 'noinline' ) {
+			$noinline = 1;
+		    } elsif ( $_ eq 'nolog' ) {
+			$nolog = 1;
+		    } elsif ( $_ eq 'builtin' ) {
+			$builtin = 1;
+		    } else {
+			fatal_error "Invalid option ($_)";
+		    }
+		}
+	    }
+
+	    fatal_error "Conflicting OPTIONS ($options)" if $noinline && $type == INLINE;
+
+	    if ( my $actiontype = $targets{$action} ) {
+		if ( ( $actiontype & ACTION ) && ( $type == INLINE ) ) {
+		    if ( $actions{$action}->{noinline} ) {
+			warning_message "'inline' option ignored on action $action -- that action may not be in-lined";
+			next;
+		    }
+		    
+		    delete $actions{$action};
+		    delete $targets{$action};
+		} else {
+		    warning_message "Duplicate Action Name ($action) Ignored" unless $actiontype & ( ACTION | INLINE );
+		    next;
+		}
+	    }
+
+	    if ( $builtin ) {
+		$targets{$action}         = USERBUILTIN + OPTIONS;
+		$builtin_target{$action}  = 1;
+	    } else {
+		new_action $action, $type, $noinline, $nolog;
+
+		my $actionfile = find_file( "action.$action" );
+
+		fatal_error "Missing Action File ($actionfile)" unless -f $actionfile;
+
+		$inlines{$action} = { file => $actionfile, nolog => $nolog } if $type == INLINE;
+	    }
+	}
+    }
+
+    if ( my $action = $config{REJECT_ACTION} ) {
+	my $type = $targets{$action};
+	fatal_error "REJECT_ACTION ($action) was not defined"  unless $type;
+	fatal_error "REJECT_ACTION ($action) is not an action" unless $type == INLINE;
+    }
+}
+
 #
 # Create a policy action if it doesn't already exist
 #
@@ -1818,6 +1830,41 @@ sub use_policy_action( $$ ) {
     $ref;
 }
 
+#
+# Process the REJECT_ACTION
+#
+sub process_reject_action() {
+    my $rejectref = $filter_table->{reject};
+    my $action    = $config{REJECT_ACTION};
+
+    if ( ( $targets{$action} || 0 ) == ACTION ) {
+	add_ijump $rejectref, j => use_policy_action( $action, $rejectref->{name} );
+    } else {
+	process_inline( $action,      #Inline
+			$rejectref,   #Chain
+			'',           #Matches
+			'',           #Log Level and Tag
+			$action,      #Target
+		        '',           #Param
+			'-',          #Source
+			'-',          #Dest
+			'-',          #Proto
+			'-',          #Ports
+			'-',          #Sports
+			'-',          #Original Dest
+			'-',          #Rate
+			'-',          #User
+			'-',          #Mark
+			'-',          #ConnLimit
+			'-',          #Time
+			'-',          #Headers
+			'-',          #Condition
+			'-',          #Helper
+			0,            #Wildcard
+	    );
+    }
+}
+
 ################################################################################
 # End of functions moved from the Actions module in 4.4.16
 ################################################################################
@@ -2002,7 +2049,7 @@ sub process_inline ($$$$$$$$$$$$$$$$$$$$$) {
 
 	my $actiontype = $targets{$action} || find_macro( $action );
 
-	fatal_error( "Invalid Action ($mtarget) in inline action" ) unless $actiontype & ( ACTION + STANDARD + NATRULE + MACRO + CHAIN + INLINE );
+	fatal_error( "Invalid Action ($mtarget) in inline action" ) unless $actiontype & ( ACTION + STANDARD + NATRULE + MACRO + CHAIN + INLINE + INLINERULE );
 
 	if ( $msource ) {
 	    if ( $msource eq '-' ) {
@@ -2127,11 +2174,11 @@ sub process_rule ( $$$$$$$$$$$$$$$$$$$ ) {
     if ( $basictarget eq 'INLINE' ) {
 	my $inline_matches = get_inline_matches;
 
-	if ( $inline_matches =~ /^(.*\s+)-j\s+(.+) $/ ) {
-	    $raw_matches .= $1;
+	if ( $inline_matches =~ /^(.*\s+)?-j\s+(.+) $/ ) {
+	    $raw_matches .= $1 if supplied $1;
 	    $action = $2;
 	    my ( $target ) = split ' ', $action;
-	    fatal_error "Unknown jump target ($action)" unless $targets{$target};
+	    fatal_error "Unknown jump target ($action)" unless $targets{$target} || $target eq 'MARK';
 	    fatal_error "INLINE may not have a parameter when '-j' is specified in the free-form area" if $param ne '';
 	} else {
 	    $raw_matches .= $inline_matches;
@@ -2428,7 +2475,7 @@ sub process_rule ( $$$$$$$$$$$$$$$$$$$ ) {
 		fatal_error "Rules may not override a NONE policy";
 	    }
 	    #
-	    # Handle Optimization
+	    # Handle Optimization level 1 when specified alone
 	    #
 	    if ( $optimize == 1 && $section == NEW_SECTION ) {
 		my $loglevel = $filter_table->{$chainref->{policychain}}{loglevel};
@@ -2450,10 +2497,12 @@ sub process_rule ( $$$$$$$$$$$$$$$$$$$ ) {
 		my $auxref   = $filter_table->{$auxchain};
 
 		unless ( $auxref ) {
+		    my $save_comment = push_comment;
 		    $auxref = new_chain 'filter', $auxchain;
 		    $auxref->{blacklistsection} = 1 if $blacklist;
 
 		    add_ijump( $chainref, j => $auxref, state_imatch( $section_states{$section} ) );
+		    pop_comment( $save_comment );
 		}
 
 		$chain    = $auxchain;
@@ -2609,7 +2658,7 @@ sub process_rule ( $$$$$$$$$$$$$$$$$$$ ) {
 	     }
 
 	fatal_error "$basictarget rules are not allowed in the $section_rmap{$section} SECTION" if $actiontype & ( NATRULE | NONAT );
-	$rule .= "$globals{STATEMATCH} ESTABLISHED " if $section == ESTABLISHED_SECTION;
+	$rule .= state_match('ESTABLISHED') if $section == ESTABLISHED_SECTION;
     }
     #
     # Generate CT rules(s), if any
@@ -2850,6 +2899,8 @@ sub perl_action_helper($$;$) {
 
     $matches .= ' ' unless $matches =~ /^(?:.+\s)?$/;
 
+    set_inline_matches $matches if $target =~ /^INLINE(?::.*)?$/;
+
     if ( $isstatematch ) {
 	if ( $statematch ) {
 	    if ( $statematch eq $isstatematch ) {
