@@ -26,7 +26,7 @@
 #       You may only use this script to uninstall the version
 #       shown below. Simply run this script to remove Shorewall Firewall
 
-VERSION=4.5.18
+VERSION=4.5.21.10
 
 usage() # $1 = exit status
 {
@@ -116,8 +116,28 @@ fi
 
 rm -f ${SBINDIR}/shorewall
 
+if [ -f "$INITSCRIPT" ]; then
+    if mywhich updaterc.d ; then
+	updaterc.d ${PRODUCT} remove
+    elif mywhich insserv ; then
+        insserv -r $INITSCRIPT
+    elif mywhich chkconfig ; then
+	chkconfig --del $(basename $INITSCRIPT)
+    elif mywhich systemctl ; then
+	systemctl disable ${PRODUCT}
+    fi
+
+    remove_file $INITSCRIPT
+fi
+
 rm -rf ${SHAREDIR}/shorewall/version
 rm -rf ${CONFDIR}/shorewall
+
+if [ -n "$SYSCONFDIR" ]; then
+    [ -n "$SYSCONFFILE" ] || SYSCONFFILE=${PRODUCT};
+    rm -f ${SYSCONFDIR}/${SYSCONFFILE}
+fi
+
 rm -rf ${VARDIR}/shorewall
 rm -rf ${PERLLIB}/Shorewall/*
 rm -rf ${LIBEXEC}/shorewall
